/* fuga.c
 *
 * author marco corvi ( marco_corvi@geocities.yahoo.com )
 * date
 *
 *
 * Vector-International Fuga Camera
 * http://www.vector-international.be/C-Cam/PCI_card/PCIDOC.html
 *
 * credits (ie, copied from)
 * PCI.C source code, thanks to Dr. Boelen <josy@vector-international.be>
 * PCIGNU.C (from Vector-Int'l website)
 * the linux fuga driver by Tim Llewellynn <tgl13@elec.canterbury.ac.nz>
 *     http://www.elec.canterbury.ac.nz/research/3d_vision/linux/drivers.html
 *
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif

#ifndef MODULE
#  define MODULE
#endif

#define __NO_VERSION__ /* don't define kernel_verion in module.h */
#include <linux/module.h>
#include <linux/version.h>
char kernel_version [] = UTS_RELEASE;

#define PRINTE( fmt, args...) printk(KERN_EMERG fmt, ##args)
#define PRINTA( fmt, args...) printk(KERN_ALERT fmt, ##args)
#define PRINTC( fmt, args...) printk(KERN_CRIT  fmt, ##args)
#define PRINTR( fmt, args...) printk(KERN_ERR   fmt, ##args)
#define PRINTW( fmt, args...) printk(KERN_WARNING fmt, ##args)
#define PRINTN( fmt, args...) printk(KERN_NOTICE fmt, ##args)
#define PRINTI( fmt, args...) printk(KERN_INFO  fmt, ##args)
#define PRINTD( fmt, args...) printk(KERN_DEBUG fmt, ##args)

#define PRINTZ( fmt, args...) /* nothing */

#include <linux/sched.h>
#include <linux/delay.h>   /* udelay */
#include <linux/fs.h>  /* struct file_operations */
#include <linux/mm.h>  
#include <linux/stat.h>
#include <linux/file.h>
#include <linux/malloc.h> /* kmalloc kfree */
#include <asm/uaccess.h>  /* VERIFY_* */
#include <asm/io.h>       /* in# out# */

/* #include <linux/bios32.h> obsolete: see Documentation/pci.txt */
#include <linux/pci.h>
#include "fuga.h"
#include "fuga_def.h"
#include "fuga_ioctl.h"
/*
 * Refuse to compile under versions older than 1.99.4
 */
#define VERSION_CODE(vers,rel,seq) ( ((vers)<<16) | ((rel)<<8) | (seq) )
#if LINUX_VERSION_CODE < VERSION_CODE(1,99,4)
#  error "This module needs Linux 1.99.4 or newer"
#else

/*
 * maximum number of tries before breaking out of a loop 
 */
#define FUGA_MAX_TRY 10000000

/*
 * PCI hardware addresses
 */
static u32 addresses[] = {
    PCI_BASE_ADDRESS_0,
    PCI_BASE_ADDRESS_1,
    PCI_BASE_ADDRESS_2,
    PCI_BASE_ADDRESS_3,
    PCI_BASE_ADDRESS_4,
    PCI_BASE_ADDRESS_5,
    0
}; 

static struct fuga_amcc fuga5920 = {
  0x10e8, 0x5920, 0x00, 0x90, {0x00, 0x00}, 
  0x01, {0x00, 0x80, 0x04}, 0x00, 0xe0, 0x00, 0x00,
  { 0x10e8ffc1, 0xbffffffd, 
  0xbffffffd, 0xffffff01,
  0x00000000, 0x00000000 }, 
  {'V', 'e', 'c', 't', 'o', 'r', 0x00, 0x00}, 0x00000000,
  {'I', 'n', 't', 'l', 0x00, 0x00, 0x00, 0x00}, 0xff, 0x01, 0x00, 0x00
};

static u8 fuga_buffer[FUGA_BUFFER_SIZE];
static struct fuga_device * fuga_dev;

/* ###############################################################
 * UTILITY ROUTINES
 *     fuga_ccam_reset
 *     fuga_info()
 * ###############################################################
 */

/* ---------------------------------------------------------------
 * fuga_ccam_reset 
 *
 * driver soft reset to the PCI CCam card default values
 * called by - ioctl IOC_RESET
 *           - init_module
 *           - ??? open
 *
 * return 0 (always)
 * ---------------------------------------------------------------
 */
int fuga_ccam_reset( struct fuga_device * dev )
{
  u8 camera = 0;
  struct fuga_ccam * ccam = &(dev->ccam);

  ccam->flags = FUGA_CORRECTION_FLAG | (camera & FUGA_CAMERA_FLAGS);
  /* 
   * default camera = 0;   offset correction = 1; 
   * calibration    = 0;   loopback = 0;
   */
  ccam->ystart   = 0;
  ccam->yend     = 511;
  ccam->yinc     = 1;
  ccam->xstart   = 0;
  ccam->xend     = 511; 
  ccam->xinc     = 1;
  ccam->vdelay   = 0x10; /* vertical delay : was 0x80 */
  ccam->hdelay   = 0x10; /* horizontal delay */
  ccam->xhdelay  = 0x10; /* horizontal extra delay (every 32 cols) */
  ccam->adcdelay = 0x03;
  ccam->gain     = 80;   /* default external gain control */
  /* ccam->xaddr; */
  /* ccam->yaddr; */
  return 0;
}

/* ---------------------------------------------------------------
 * fuga_ccam_validate
 *
 * return 0   ok
 *        -1  invalid W.o.I. (eg, ystart>yend)
 * ---------------------------------------------------------------
 */
int fuga_ccam_validate( struct fuga_device * dev )
{ 
  struct fuga_ccam * ccam = &(dev->ccam);

  ccam->ystart &= FUGA_START_MASK;
  ccam->xstart &= FUGA_START_MASK;
  ccam->yinc &= FUGA_INC_MASK;
  ccam->xinc &= FUGA_INC_MASK;
  ccam->yend &= FUGA_END_MASK;
  ccam->xend &= FUGA_END_MASK;

  if (ccam->yinc == 0) ccam->yinc = 1;
  if (ccam->xinc == 0) ccam->xinc = 1;

  ccam->vdelay &= FUGA_VDELAY_MASK;
  ccam->hdelay &= FUGA_HDELAY_MASK;
  ccam->xhdelay &= FUGA_HDELAY_MASK;

  ccam->adcdelay &= FUGA_ADCDELAY_MASK;
  if (ccam->adcdelay < 3) ccam->adcdelay = 3;

  ccam->gain &= FUGA_GAIN_MASK;

  ccam->flags &= FUGA_FLAGS_MASK;

  if (ccam->ystart > ccam->yend) 
    return -1;
  if (ccam->xstart > ccam->xend) 
    return -1;
  
  return 0;
}

/* ##################################################################
 * NOVRAM ROUTINE
 *     fuga_novram_write
 * #################################################################
 */

/* ------------------------------------------------------------------
 * fuga_novram_write
 *
 * write the amcc data to novram (i/o region 0 )
 * called by - ioctl IOCS_AMCC
 *
 * returns  0    ok
 *          -EIO write error
 *
 * N.B. This routine has not been tested
 * ------------------------------------------------------------------
 */
int fuga_novram_write( struct fuga_device * dev)
{
#ifdef Gfgfkskhahadsjhajs
  u32 base = dev->amcc.address[0];
  u8 * s;
  u8 b;
  int i;
  u16 nvcmd, nvdata;

  nvcmd = (u16) (base + 0x3f);
  nvdata = (u16) (base + 0x3e);

  s = (u8 *)&(dev->amcc);
  for (i=0; i<sizeof(struct fuga_amcc); i++) {
    unsigned long jif_start = jiffies;
    while ( ( (b = (u8)inb(nvcmd)) & 0x80) != 0) { 
      PRINTE( "wait on nvcmd ");
      udelay(10);
      schedule(); 
    }
    PRINTD("fuga_novram_write waited %ld jiffies\n", (jiffies - jif_start) );
    outb( 0x80, nvcmd );     /* load low address */
    mdelay( 100 );
    outb( 0x40+i, nvdata );

    outb( 0xa0, nvcmd );    /* load high address */
    mdelay( 100 );
    outb( 0x00, nvdata );

    outb( 0x00, nvcmd );       /* set inactive */
    mdelay( 100 );
    outb( *s, nvdata );

    outb( 0xc0, nvcmd );    /* write the byte */
    mdelay( 100 );
    do { } while ( (inb(nvcmd) & 0x80) != 0);

    outb( 0x80, nvcmd );
    mdelay( 100 );
    outb( 0x40+i, nvdata );    /* low address */ 

    outb( 0xe0, nvcmd );
    mdelay( 100 );
    do { } while ( (inb(nvcmd) & 0x80) != 0);

    if ( ( b = (u8) inb( nvdata ) ) != (u8) *s ) {
      PRINTE( "fuga_novram_write verify error at %d: %x-%x\n", i, b, *s); 
      return (-EIO);
    }
    ++ s;
  }
#endif
  PRINTE("fuga_write_novram done\n");
  return 0;
}

/* ################################################################
 * FLEX ROUTINES
 *    fuga_flex_config
 *    fuga_flex_ttf
 * ################################################################
 */

/* ----------------------------------------------------------------
 * fuga_flex_config
 *
 * set/clear the config line to the FLEX ???????
 * called by - ioctl IOC_CONFIG
 *
 * returns 0 (always)
 * ----------------------------------------------------------------
 */
int fuga_flex_config( struct fuga_device * dev)   /* case 'c' */ 
{
  u8 nbr1, nbr2;
  u32 base1 = dev->amcc.address[1];

  nbr1 = (inb_p(base1) & 2) == 0;          /* ???? */
  nbr1 = ( (inb_p(base1)>>1) & 1) +0x30 ;  /* ???? 0x30='0' for display */
  outb_p( (nbr1 != 0)? 1: 0, base1 );
  nbr2 = ( (inb_p(base1)>>1) & 1) +0x30 ;  /* ???? read and display ? */
  
  return 0;
}


/* ----------------------------------------------------------------
 * fuga_flex_ttf
 * 
 * upload ttf file (buffer) in the flash
 * called by - ioctl FUGA_IOCS_TTF
 *
 * return 0       o.k.
 *        -EACCES verify read failure
 *        -ETIME  status low on start 
 *        -EIO    status low after writing
 * ----------------------------------------------------------------
 */
int fuga_flex_ttf( struct fuga_device * dev, u8 * buf) /* cases 'g' and 'f' */
{
  u32 base1 = ((dev->amcc).address)[1];   /* config io port */
  u32 base2 = ((dev->amcc).address)[2];   /* status io port */
  u8 c1,c2,c3,c4, c1b,c2b,c3b,c4b;
  int i, k=0; 
  volatile u8 c;

  if (access_ok(VERIFY_READ, buf, FUGA_TTF_SIZE) != 1) 
    return (-EACCES);

  c1b = inb( base2 );    c1  = inb( base1 );
  outb( 0+2, base1 );                         /* lower config */
  c2b = inb( base2 );    c2  = inb( base1 );
  udelay( 250 );
  outb( 1+2, base1 );                         /* raise config */
  c3b = inb( base2 );    c3  = inb( base1 );
  udelay( 500 );                              /* check again */
  c4b = inb( base2 );    c4  = inb( base1 );

  for (i=0; i<1000; i++) {   /* wait for nStatus to go up ... */
    c = inb_p( base2 );
    if ( (c & 0x01) == 1) break;
    udelay( 100 );
  } 
  if (i == 1000)                        /* status low on start */
    return (-ETIME);
    
  for (i=0; i<FUGA_TTF_SIZE; i++) {
    if ( ( i % 512) == 0) {
      copy_from_user(fuga_buffer, buf+i, 
         ( (FUGA_TTF_SIZE-i) > FUGA_BUFFER_SIZE) ? 
           FUGA_BUFFER_SIZE : FUGA_TTF_SIZE-i);
      k = 0;   /* reset at the beginning of the buffer */ 
    }
    /* 
     * simply use a byte ....
     */
    c = (*(fuga_buffer+k) );
    c1 = c;
    outb( c, base2);  c >>= 1;
    outb( c, base2);  c >>= 1;
    outb( c, base2);  c >>= 1;
    outb( c, base2);  c >>= 1;
    outb( c, base2);  c >>= 1;
    outb( c, base2);  c >>= 1;
    outb( c, base2);  c >>= 1;
    outb( c, base2);  
    c2 = inb( base2 );
    if ((c2 & 0x01)==0) {
      PRINTD("fuga_flex_ttf status %02x low at byte %d %02x %02x\n", 
         c2, i, c1, c);
      return (-EIO);
    }
    k++;
  }
  for (k=0; k<10; k++) {  /* ten more DCLKS */
    outb( c, base2 );  
  }
  return 0;
}

/* ###################################################################
 * FLASH ROUTINES
 *     fuga_flash_write()
 *     fuga_flash_clear()
 *     fuga_flash_read()
 *     fuga_flash_test()
 *     fuga_flash_status()
 *     fuga_flash_erase()
 *     fuga_flash_reset()
 *     fuga_flash_byte_write()
 *     fuga_flash_byte_read()
 * ###################################################################
 */

/* ----------------------------------------------------------------
 * fuga_flash_write 
 *
 * upload the camera image (data from a NIP file) into the flash.
 * The NIP file should be read from the camera without CALIBRATION_FLAG
 * The file has a six-byte header with:
 *        'B', y-size (2 bytes), x-size (2 bytes), 1 (B/W) or 3 (color)
 * The header is followed by the stream of pixels uchars.
 * The pixel values are offset by 128 in the NIP format, while they are
 * unchanged in the BIN format.
 * 
 * The image data in "buf" must have size 512x512
 *
 * return number of byte written :
 *               262144 on success, less on error
 *        -EACCES   verify read failure
 *        ...       flash_reset error
 *        -status   flash_erase error
 * ----------------------------------------------------------------
 */
int fuga_flash_write( struct fuga_device * dev, u8 * buf)
{
  int ret;
  char * k;
  u16 x, y;

  if ( access_ok(VERIFY_READ, buf, FUGA_BUFFER_SIZE*FUGA_BUFFER_SIZE) != 1) {
    return (-EACCES);
  }

  if ( (ret = fuga_flash_reset( dev )) < 0) {
    PRINTZ("fuga_flash_write reset fail\n");
    return ret;        /* clear flash from any startup error */
  }
  /*
   * do not require status+FUGA_FLASH_STATUS_READY
   * otherwise i do not get through
   */
  if ( (ret = fuga_flash_erase( dev, 0x000000 ) ) < 0) {
    PRINTZ("<0> fuga_flash_write erase 0 fail status = %04x\n", ret);
    return -ret; 
  }
  if ( (ret = fuga_flash_erase( dev, 0x010000 ) ) < 0) {
    PRINTZ("<0> fuga_flash_write erase 1 fail status = %04x\n", ret);
    return -ret; 
  }
  if ( (ret = fuga_flash_erase( dev, 0x020000 ) ) < 0) {
    PRINTZ("<0> fuga_flash_write erase 2 fail status = %04x\n", ret);
    return -ret; 
  }
  if ( (ret = fuga_flash_erase( dev, 0x030000 ) ) < 0) {
    PRINTZ("<0> fuga_flash_write erase 3 fail status = %04x\n", ret);
    return -ret; 
  }

  printk("<0> fuga_flash_write flash status %04x \n", ret);

  for (y=0; y<FUGA_BUFFER_SIZE; y++) {
    __copy_from_user( fuga_buffer, buf+y*FUGA_BUFFER_SIZE, FUGA_BUFFER_SIZE); 
    k = fuga_buffer;
    for (x=0; x<FUGA_BUFFER_SIZE; x++) {   
      /* 
       * PCI.C had (instead of *k)
       *    (char)(128 - *k)  for NIP files (our case: fugacal.nip)
       *    (char)*k          for BIN (binary) files 
       * the transformation is supposed already done in the user program
       */
      if ( (ret = fuga_flash_byte_write( dev, x, y, *k) ) < 0) {
        PRINTZ("<0> flash_byte_write fail %04x\n", ret);
        return (y*FUGA_BUFFER_SIZE+x);
      }
      k ++;
    }
  }
  return (y*FUGA_BUFFER_SIZE);
}

/* ----------------------------------------------------------------
 * fuga_flash_clear
 *
 * clear the flash calibration image. 
 * called by - ioctl FUGA_IOC_CLEAR
 *
 * return 0         success
 *        ...       flash_reset error
 *        -status   flash_erase error
 *        flash_byte_write error code
 * ----------------------------------------------------------------
 */

int fuga_flash_clear( struct fuga_device * dev)  
{
  int ret;
  u16 x, y;
  u32 k = 0;

  if ( (ret = fuga_flash_reset( dev )) < 0) 
    return ret;    
  /*
   * as in fuga_flash_write do not require status=FUGA_FLASH_STATUS_READY
   */
  if ( (ret = fuga_flash_erase( dev, 0x000000 ) ) < 0) 
    return -ret;
  if ( (ret = fuga_flash_erase( dev, 0x010000 ) ) < 0) 
    return -ret;
  if ( (ret = fuga_flash_erase( dev, 0x020000 ) ) < 0) 
    return -ret;
  if ( (ret = fuga_flash_erase( dev, 0x030000 ) ) < 0) 
    return -ret;

  for (y=0; y<FUGA_BUFFER_SIZE; y++) {
    for (x=0; x<FUGA_BUFFER_SIZE; x++) {
      /*
       * why 0x55 if x==1 or y==1 ???
       *
      if ( (ret = fuga_flash_byte_write( dev, x, y, 
              (char)( ( k % 0x201 ) == 0) ? 0x80 : 0x55 )) != 0) {
       *
       */
      if ( (ret = fuga_flash_byte_write( dev, x, y, (char)(0x80) ) ) != 0) {
        return (y*FUGA_BUFFER_SIZE+x);
      }
      k ++;
    }
  }
  return (y*FUGA_BUFFER_SIZE);
}

/* ----------------------------------------------------------------
 * fuga_flash_read
 *
 * read a 512x512 camera image into the user buffer "buf"
 * (used to create a calibration NIP or BIN  file)
 *
 * return number of byte read
 *        -EACCES verify write failure
 *
 * N.B. bytes are copyed to user program in bunch of FUGA_BUFFER_SIZE (512)
 *      therefore only n-(n%512) are made available in case of error
 * ----------------------------------------------------------------
 */
int fuga_flash_read( struct fuga_device * dev, u8 * buf)
{ int n;
  u16 x, y;
  u8 * k;

  if (access_ok(VERIFY_WRITE, buf, FUGA_BUFFER_SIZE*FUGA_BUFFER_SIZE) != 1) 
    return (-EACCES);

  for (y=0; y<FUGA_BUFFER_SIZE; y++) {
    k = fuga_buffer;
    for (x=0; x<FUGA_BUFFER_SIZE; x++) {
      if ( (n = fuga_flash_byte_read( dev, x, y )) < 0) 
        return (y*FUGA_BUFFER_SIZE+x);
      *k++ = (unsigned char) n;
    }
    __copy_to_user( buf+y*FUGA_BUFFER_SIZE, fuga_buffer, FUGA_BUFFER_SIZE);
  }
  return (y*FUGA_BUFFER_SIZE);
}

/* ----------------------------------------------------------------
 * fuga_flash_test
 *
 * test the flash i/o
 * why do we want this routine ? after all it is just a write 
 * followed by a read ...
 *
 * called by ioctl FUGA_IOC_TEST
 *
 * return byte number (262144) o.k.
 *        -byte_number writing error (flash_byte_write timeout)
 *        +byte_number reading error (flash_byte_read timeout)
 *        ...       flash_reset error
 *        -status   flash_erase error (timeout)
 *        -EIO      flash_byte_write error
 *        -EINVAL   byte check error
 * ----------------------------------------------------------------
 */
int fuga_flash_test( struct fuga_device * dev) /* case 'j' */
{
  int ret = 0;
  u16 x, y;
  u8 k;
  

  if ( ( ret = fuga_flash_reset( dev )) < 0) {
    PRINTZ("test reset fail \n");
    return ret;
  }

  /*
   * as in fuga_flash_write do not require status=FUGA_FLASH_STATUS_READY
   */
  if ( (ret = fuga_flash_erase(dev, 0x00000)) < 0) 
    return (-ret);
  if ( (ret = fuga_flash_erase(dev, 0x10000)) < 0) 
    return (-ret);
  if ( (ret = fuga_flash_erase(dev, 0x20000)) < 0) 
    return (-ret);
  if ( (ret = fuga_flash_erase(dev, 0x30000)) < 0) 
    return (-ret);
  
  printk("<0> fuga_flash_test write\n");

  k = 0;
  for (y=0; y<FUGA_BUFFER_SIZE; y++) {
    for (x=0; x<FUGA_BUFFER_SIZE; x++) {
      if ((x%256) == 0) {
        if (k<255) { k++; } else { k=0; }
      }
      if ( (ret=fuga_flash_byte_write( dev, x, y, k) ) < 0) {
        printk( "<0> fuga_flash_test write error %04x at %d %d\n", ret, x, y );
        return (-y*FUGA_BUFFER_SIZE-x);
      }
      /*
       * the return stattus is not FUGA_FLASH_STATUS_READY (0x80)
       * 
      if (ret != FUGA_FLASH_STATUS_READY) {
        printk( "<0> test write status %04x at %d %d\n", ret, x, y );
        return (-EIO);
      }
       */
    }
  }
  
  k = 0; 
  for (y=0; y<FUGA_BUFFER_SIZE; y++) {
    for (x=0; x<FUGA_BUFFER_SIZE; x++) {
      if ((x%256) == 0) {
        if (k<255) { k++; } else { k=0; }
      }
      if ( (ret = fuga_flash_byte_read( dev, x, y)) < 0) 
        return (y*FUGA_BUFFER_SIZE+x);
      if ( (char)(ret) != k) {
        printk( "<0> test check error %02x (expected %02x) at %d %d\n", 
          ret, k, x, y );
        return (-EINVAL);
      }
    }
  }
  return (y*FUGA_BUFFER_SIZE);
}


/* ----------------------------------------------------------------
 * FLASH_WRITE FLASH_READ macros
 *
 * issue a WRITE (0x0020) or READ (0x0000) command 
 * and wait until the status word (offset 0x14) signal a CYCLE (0x0040)
 *
 * N.B. caller return -ETIME on timeout
 * ----------------------------------------------------------------
 */

#define FLASH_WRITE \
  outw( FUGA_CTRL_ON | FUGA_CTRL_FLASH_CYCLE | FUGA_CTRL_FLASH_WRITE, \
        base+FUGA_CTRL );  \
  { unsigned int t = FUGA_MAX_TRY; \
    while ( (inw( base + FUGA_STATUS ) & FUGA_CTRL_FLASH_CYCLE ) != 0 \
          && t > 0 ) { t --; }  \
    if (t == 0)  \
      return -ETIME; \
  } \

#define FLASH_READ \
  outw( FUGA_CTRL_ON | FUGA_CTRL_FLASH_CYCLE | FUGA_CTRL_FLASH_READ, \
        base+FUGA_CTRL );        \
  { unsigned int t = FUGA_MAX_TRY; \
    while ( (inw( base + FUGA_STATUS ) & FUGA_CTRL_FLASH_CYCLE ) != 0 \
          && t > 0 ) { t --; }  \
    if (t == 0)  \
      return -ETIME; \
  } \

/* ----------------------------------------------------------------
 * fuga_flash_status
 *
 * retrieve the flash status (within a timeout "to")
 * status is in bit 7 (0x80) in byte 0x10 of i/o region 3
 * called by - fuga_flash_byte_write
 *           - fuga_flash_erase
 *
 * return status (>=0) 
 *        neg.value if error: -EIO
 * ----------------------------------------------------------------
 */
int fuga_flash_status( struct fuga_device * dev, unsigned long to ) 
{ 
  u32 base = dev->amcc.address[3];
  int status;

  if (to == 0)  to = 0x1fffff;
  status = -1;
  do {
    FLASH_READ
            /* read status on DATA=0x10 offset */
    if ( ( (status = (int)inb(base+FUGA_FLASH_DATA)) 
           & FUGA_FLASH_STATUS_READY) != 0)      /* READY=0x80 */
      break;
    -- to;
  } while (to != 0);
  if ( (to == 0) ||
       (status & ( FUGA_FLASH_STATUS_ES |        /* ES=0x20 */
                   FUGA_FLASH_STATUS_BWS |       /* BWS=0x10 */
                   FUGA_FLASH_STATUS_VPP ) ) ) { /* VPP=0x08 */
    if (to == 0) {
      dev->timeouts ++;
/*
      if ((dev->timeouts % 0x20000) == 0)  // once every 32K timeouts //
        printk("<0> fuga_flash_status timeouts %ld status %04x\n", 
          dev->timeouts, status);
*/
    }
    outb( FUGA_FLASH_CMD_CLEAR, base + FUGA_FLASH_DATA);
    FLASH_WRITE
  }
  outb( FUGA_FLASH_CMD_NORMAL, base + FUGA_FLASH_DATA);
  FLASH_WRITE
  return ( status );
}

/* ----------------------------------------------------------------
 * fuga_flash_reset
 *
 * when should i reset the card ?
 * called by - fuga_flash_write()
 *           - fuga_flash_clear()
 *           - fuga_flash_test()
 * 
 * return 0       ok
 *        -ETIME  on FLASH_WRITE timeout
 * ----------------------------------------------------------------
 */
int fuga_flash_reset( struct fuga_device * dev ) 
{ 
  u32 base = dev->amcc.address[3];

  outb_p(FUGA_FLASH_CMD_CLEAR, base + FUGA_FLASH_DATA);  /* CLEAR=0x50 */
  FLASH_WRITE
  outb_p(FUGA_FLASH_CMD_NORMAL, base + FUGA_FLASH_DATA); /* NORMAL=0xff */
  FLASH_WRITE  

  return 0;
}

/* ----------------------------------------------------------------
 * fuga_flash_byte_write
 *
 * write the "data" to the address (= X<<9 + Y )
 * called by - fuga_flash_write
 *             fuga_flash_clear
 *
 * return status (>=0) if ok
 *        -ETIME     error (from READ/WRITE macros timeout)
 *        -EIO       in case of fuga_flash_status error
 * ----------------------------------------------------------------
 */
int fuga_flash_byte_write( struct fuga_device * dev, 
    u16 xaddr, u16 yaddr, char data) 
{
  u32 base = dev->amcc.address[3];

  /* 
   * X low - Y upper convention (see PCI.C)
   */
  outw_p( xaddr & 0x01ff, base + FUGA_FLASH_ADDRESS_L);
  outw_p( yaddr & 0x01ff, base + FUGA_FLASH_ADDRESS_U);
  outb_p( FUGA_FLASH_CMD_WRITE, base + FUGA_DATA);    /* WRITE=0x40 */
  FLASH_WRITE

  outb_p(data, base + FUGA_FLASH_DATA);   /* FUGA_FLASH_DATA=0x10 */
  FLASH_WRITE

  return ( fuga_flash_status( dev, FUGA_FLASH_WRITE_TIMEOUT ) );
}

/* ----------------------------------------------------------------
 * fuga_flash_byte_read
 *
 * read a byte from the address (= X<<9 + Y )
 * called by - fuga_flash_read
 *
 * return read byte
 *        -ETIME  in case of FLASH_READ timeout
 * ----------------------------------------------------------------
 */
int fuga_flash_byte_read( struct fuga_device * dev, u16 xaddr, u16 yaddr) 
{
  u32 base = dev->amcc.address[3];

  /* 
   * X low - Y upper convention
   */
  outw_p( xaddr & 0x01ff, base + FUGA_FLASH_ADDRESS_L);
  outw_p( yaddr & 0x01ff, base + FUGA_FLASH_ADDRESS_U);
  FLASH_READ

  return ( (int)inb( base+FUGA_FLASH_DATA ) );
}

/* ----------------------------------------------------------------
 * fuga_flash_erase
 *
 * erase flash content at the address (= X<<9 + Y) 
 * called by - fuga_flash_write
 *           - fuga_flash_clear
 *           - fuga_flash_test
 *
 * return status (>=0) if ok
 *        -ETIME for a FLASH_WRITE timeout
 * ----------------------------------------------------------------
 */
int fuga_flash_erase( struct fuga_device * dev, u32 addr)
{
  u32 base = dev->amcc.address[3];

  outw_p( (u16)(addr & 0x01ff), base + FUGA_FLASH_ADDRESS_L);
  outw_p( (u16)((addr>>9) &0x01ff), base + FUGA_FLASH_ADDRESS_U);
  outb_p( FUGA_FLASH_CMD_ERASE, base + FUGA_DATA);    /* ERASE=0x20 */
  FLASH_WRITE
  outb_p( FUGA_FLASH_CMD_CONFIRM, base + FUGA_DATA);  /* CONFIRM=0xd0 */
  FLASH_WRITE

  return ( fuga_flash_status( dev, FUGA_FLASH_ERASE_TIMEOUT ) );
}

/* ############################################################
 * FLASH UTILITY ROUTINES
 *    fuga_set_woi
 *    fuga_set_flags
 *    fuga_reset
 * ############################################################
 */

/* ----------------------------------------------------------------
 * fuga_set_woi
 *
 * writes WOI, delays, and gain to the card
 *
 * return 0 (always)
 * ----------------------------------------------------------------
 */
int fuga_set_woi( struct fuga_device * dev )
{
  u32 base = dev->amcc.address[3];
  struct fuga_ccam * ccam = &(dev->ccam);

  outw( ccam->ystart, base+FUGA_YSTART );
  outw( ccam->yinc | (ccam->adcdelay << 12), base+FUGA_YINC );
  outw( ccam->yend,   base+FUGA_YEND );
  outw( ccam->vdelay, base+FUGA_YWAIT );
  outw( ccam->xstart, base+FUGA_XSTART );
  outw( ccam->xinc,   base+FUGA_XINC );
  outw( ccam->xend,   base+FUGA_XEND );
  outw( (ccam->xhdelay<<8) | ccam->hdelay, base+FUGA_XWAIT );
  outw( ccam->gain,   base+FUGA_GAIN );

  return 0;
}

/* ----------------------------------------------------------------
 * fuga_set_flags 
 *
 * called by ioctl() IOCS_FLAGS and by read()
 *
 * return 0 (always)
 * ----------------------------------------------------------------
 */
int fuga_set_flags( struct fuga_device * dev) {
  u32 base = dev->amcc.address[3];
  struct fuga_ccam * ccam = &(dev->ccam);

  outw_p( FUGA_CTRL_CAMERA_MASK, base+FUGA_CTRL);
  outw_p( (ccam->flags & FUGA_CAMERA_FLAGS) | FUGA_CTRL_ON, base+FUGA_CTRL);
  outw_p( ((ccam->flags & FUGA_LOOPBACK_FLAG) ? FUGA_CTRL_ON : 0 ) 
          |  FUGA_CTRL_CAMERA_ADDRESS, base+FUGA_CTRL);
  outw_p( ((ccam->flags & FUGA_CORRECTION_FLAG) ? FUGA_CTRL_ON : 0)
          | FUGA_CTRL_CAMERA_CORRECTION, base+FUGA_CTRL);
  // outw_p( FUGA_CTRL_ON | FUGA_CTRL_CAMERA_CYCLE, base+FUGA_CTRL);

  return 0;
}

/* ----------------------------------------------------------------
 * fuga_reset
 *
 * called by - ioctl IOC_RESET
 *           - ioctl IOCS_TTF
 * 
 * return 0 (always)
 * ----------------------------------------------------------------
 */
int fuga_reset( struct fuga_device * dev) {
  fuga_set_woi( dev );
  fuga_set_flags( dev );
  return 0;
}


/* #################################################################
 * PCI HARDWARE
 *     fuga_hw_init()
 * #################################################################
 */

/* ----------------------------------------------------------------
 * fuga_hw_init() 
 *
 * called by init_module() to find the PCI card
 * and to read its content (addresses, and other)
 *
 * return  0        o.k.
 *        -ENODEV   no pci or no device
 * ----------------------------------------------------------------
 */
int fuga_hw_init( struct fuga_device * dev )
{
  int i;
  int bus, fun;
  unsigned char headertype;
  unsigned short vendorid, deviceid;

  dev->amcc = fuga5920; /* struct copy / default amcc config*/

  if (!pcibios_present()) 
    return (-ENODEV);
  
  bus=0; /* only bus 0 :-) */
  for (fun=0; fun < 0x100; fun++) {
    headertype = 0;
    if (!PCI_FUNC(fun)) /* first function */
      pcibios_read_config_byte(bus,fun,PCI_HEADER_TYPE,&headertype);
    else if (!(headertype&0x80))
      continue;
    pcibios_read_config_word(bus,fun,PCI_VENDOR_ID, &vendorid);
    if (vendorid == FUGA_VENDOR) {
      /* A device with the correct vendor id was found */
      pcibios_read_config_word(bus,fun,PCI_DEVICE_ID,&deviceid);

      dev->amcc.vendor_id = vendorid;
      dev->amcc.device_id = deviceid;
      dev->bus_nr = 0;
      dev->fun_nr = fun;

      /* 
       * read the four addresses
       */
      for (i=0; addresses[i]; i++) {
        pcibios_read_config_dword(bus, fun, addresses[i],
          &(((dev->amcc).address)[i]) );  
        ((dev->amcc).address)[i] &= 0xfffffff8;  /* mask bits 2-0 */
        /*
         * bit-0 is   1=i/o, 0=memory
         * bits 1-2 are the type-bits (???)
         * other bits might be unused (???)
         */ 
      }
      /*
       * anything else to read ?
       * rev_id, irq, pin, latency, min_grant, max_lat, header_type, master
       */
      return 0;
    }
  }
  return -ENODEV;
}

/* #####################################################################
 * FILE OPS
 *      fuga_open()
 *      fuga_release()
 *      fuga_read()
 *      fuga_ioctl()
 * #####################################################################
 */

/* ----------------------------------------------------------------
 * fuga_open
 *
 * called by sys_open() when the user process opens the device
 * N.B. The first user process that opens the device should initialize
 *      the flex (fuga_flex_ttf)
 *
 * return  0       o.k.
 *        -EINVAL  invalid access mode
 *        -EBUSY   device already open
 * ----------------------------------------------------------------
 */
int fuga_open( struct inode *inode, struct file * filp) 
{
  if ( (filp->f_flags & O_ACCMODE) == O_WRONLY) 
    return -EINVAL;

  if (fuga_dev->use_count > 0) 
    return -EBUSY; 

  filp->private_data = (void *)fuga_dev;
  fuga_dev->use_count ++;
  fuga_dev->timeouts = 0; /* reset timeouts count */
  /*
   * no soft reset: get the device in the state it has been left 
   fuga_ccam_reset( fuga_dev );
   */
  MOD_INC_USE_COUNT;

  return 0;
}

/* ----------------------------------------------------------------
 * fuga_release
 *
 * called by sys_close() when the user process closes the device
 *
 * return 0 (always)
 * ----------------------------------------------------------------
 */
int fuga_release(struct inode *inode, struct file *filp) 
{
  struct fuga_device * dev;
  dev = (struct fuga_device *)(filp->private_data);
  if (dev->use_count == 0) {
    PRINTZ("fuga_release use count %d\n", dev->use_count);
    /* do not handle the error */
  } else {
    dev->use_count --;
  }
  MOD_DEC_USE_COUNT;
  return 0;
}

/* ------------------------------------------------------------------
 * fuga_read
 *
 * read routine: reads a region at once
 * 
 * return number of bytes read
 *        -EACCES  verify write failure
 * ------------------------------------------------------------------
 */
ssize_t fuga_read(struct file *filp, char * buf, size_t cnt, loff_t * off)
{
  struct fuga_device * dev = (struct fuga_device *)(filp->private_data);
  struct fuga_ccam * ccam = &(dev->ccam);
  u32 base = dev->amcc.address[3];
  u32 dbase = base + FUGA_DATA;
  u32 sbase = base + FUGA_STATUS;
  int status;
  int x, y;
  unsigned int n;
  u32 * lhp;
  int xcount = (ccam->xend - ccam->xstart + ccam->xinc) / ccam->xinc;
  int ye = (ccam->yend - ccam->ystart + ccam->yinc) / ccam->xinc;
  int xe = (xcount + 3)/4;
  unsigned int t;

  if ( access_ok(VERIFY_WRITE, buf, cnt) != 1) 
    return -EACCES;
  /* 
   * ystart and xstart are lost at each read
   */
  outw( ccam->ystart, base+FUGA_YSTART );
  outw( ccam->xstart, base+FUGA_XSTART );

  /* 
   * abort current cycle (if any) and restart new cycle 
   */
  outw( 0 | FUGA_CTRL_CAMERA_CYCLE, base+FUGA_CTRL);
  outw( FUGA_CTRL_ON | FUGA_CTRL_CAMERA_CYCLE, base+FUGA_CTRL);

  n = 0;                         /* how many bytes we have read so far */
  for (y=0; y<ye; y++) {         /* read one line at a time */
    lhp = (u32*)fuga_buffer;
    for (x=0; x<xe; x++) {
      for (t=FUGA_MAX_TRY; t>0; t--) {
        if ( inw(sbase) & FUGA_STATUS_CAMERA_XDRDY) {
          *lhp++ = inl( dbase );
          goto doNext;
        }
      }
      return (y*xcount+x*4);     /* read error */
doNext:
      // continue; //
    }

    if (n+xcount >= cnt) {
      copy_to_user( buf+n, fuga_buffer, cnt-n);
      n = cnt;
      break;
    } else {
      copy_to_user( buf+n, fuga_buffer, xcount);
      n += xcount;
    }
  }
  if ( (status = inw( base + FUGA_STATUS )) & FUGA_CTRL_CAMERA_CYCLE) {
    PRINTD("fuga_read state error after a read\n");
    /*
     * try to recover
     */
    outw_p( 0 | FUGA_CTRL_CAMERA_CYCLE, base+FUGA_CTRL);
  }
  status = inl( base + FUGA_DATA );
  outw( FUGA_STATUS_CAMERA_XDRDY, base + FUGA_CTRL );
  return n;
}

/* -----------------------------------------------------------------
 * fuga_ioctl
 *
 * return 
 *        -EACCES  verify read/write failure
 *        -ETIME   timeout error (PIXEL)
 *        -EPERM   not root
 * -----------------------------------------------------------------
 */
int fuga_ioctl(
  struct inode*inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
  struct fuga_device * dev = filp->private_data;
  struct fuga_ccam * ccam = &(dev->ccam);
  u32 base = dev->amcc.address[3];   
  int retval = 0;
  /* unsigned int  ui; */
  unsigned char uc;
  /* unsigned short us; */ 
  unsigned long ul;
  int size = _IOC_SIZE(cmd);
  int dir  = _IOC_DIR(cmd);
  unsigned int t;

  /*
   * check permissions 
   */ 
  if ( ( cmd == FUGA_IOCS_CALIBRATION || cmd == FUGA_IOCS_AMCC ||
         cmd == FUGA_IOC_HW_RESET     || cmd == FUGA_IOC_CONFIG ||
         cmd == FUGA_IOC_CLEAR        || cmd == FUGA_IOC_TEST )  &&
       (! suser() ) ) {
    printk("<0> fuga_ioctl perm error\n");
    return -EPERM;
  }

  /*
   * check access
   */
  if (dir & _IOC_READ)
    if (access_ok( VERIFY_WRITE, (void *)arg, size) != 1)
      return  -EACCES;
  else if (dir & _IOC_WRITE)
    if (access_ok( VERIFY_READ, (void *)arg, size) != 1)
      return -EACCES;

  retval = -EINVAL;

  switch (cmd) {
    /* -------------------------------------------------------------
     * ioctl GET :
     *    CCAM   Window Of Interest, delays, flags, gain
     *    AMCC   Camera data
     *    CALIBRATION   camera calibration image
     *    PIXEL  read a single pixel
     * -------------------------------------------------------------
     */
    case FUGA_IOCG_CCAM:
      retval = __copy_to_user( (struct fuga_ccam *)arg, ccam, 
                 sizeof(struct fuga_ccam) );
      break;
    case FUGA_IOCG_AMCC:
      retval = __copy_to_user((struct fuga_amcc *)arg, &(dev->amcc), 
                 sizeof(struct fuga_amcc) );
      break;
    case FUGA_IOCG_CALIBRATION:
      retval = fuga_flash_read( dev, (u8 *) arg);
      break;
    case FUGA_IOCG_PIXEL:
      if ( (retval = __get_user(ul, (u32 *)arg)) == 0) {
        /*
         * dev->xaddr = ul & FUGA_ADDR_MASK;
         * dev->yaddr = (ul >> 9) & FUGA_ADDR_MASK;
         */
        outl_p( ul & 0x0003ffff, dev->amcc.address[3] + FUGA_DPA);
         
        for (t=FUGA_MAX_TRY; t>0; t--) {
          retval = inw_p( dev->amcc.address[3] + FUGA_DPA);
          if ((retval & FUGA_DPA_BUSY) == 0) break;
        }
        retval = (t == 0)? -ETIME: retval & 0x00ff;
        __put_user(retval, (u32 *)arg);
      }
      break;

    /* ------------------------------------------------------------
     * ioctl SET :
     *    CCAM   Window Of Interest, delays, flags, gain
     *    AMCC   (root only) writes the novram
     *    CTRL   writes the controls COLOR/TELE/SNEL
     *    TTF    initialize camera
     *    CALIBRATION (root only) copy the calibration to the board
     * -----------------------------------------------------------
     */ 
    case FUGA_IOCS_CCAM:
      if ( (retval = __copy_from_user( ccam, (struct fuga_ccam *)arg, 
                sizeof(struct fuga_ccam)) ) == 0) {
        fuga_ccam_validate( dev ); 
        fuga_reset( dev );
      }
      break;
    case FUGA_IOCS_CALIBRATION:
      retval = fuga_flash_write( dev, (u8 *) arg);
      break;
    case FUGA_IOCS_AMCC:
      if ( (retval = __copy_from_user( &(dev->amcc), (struct fuga_amcc *)arg, 
                   sizeof(struct fuga_amcc) )) == 0) {
        retval = fuga_novram_write( dev );
      }
      break;

    case FUGA_IOCS_CTRL:
      if ( (retval = __get_user(uc, (u8 *)arg)) == 0) {
        outw( ((uc & FUGA_COLOR)? FUGA_CTRL_ON : 0) | FUGA_CTRL_COLOR,
              base+FUGA_CTRL);
        outw( ((uc & FUGA_TELE) ? FUGA_CTRL_ON: 0) | FUGA_CTRL_TELE,
              base+FUGA_CTRL);
        outw( ((uc & FUGA_SNEL) ? FUGA_CTRL_ON : 0 ) | FUGA_CTRL_SNEL, 
              base+FUGA_CTRL);
        outw( FUGA_CTRL_ON | FUGA_CTRL_CONFIG, base+FUGA_CTRL);
      }
      break;

    case FUGA_IOCS_TTF:
      if ( (retval = fuga_flex_ttf( dev, (u8 *)arg )) == 0) {
        retval = fuga_reset( dev );      // write woi + flags
      }
      break;

    /* -----------------------------------------------------------------
     * ioctl CMD :
     *    RESET,    reset default values
     *    HW_RESET  (root only)
     *    CONFIG,   (root only) clr/set flex config line
     *    CLEAR,    (root only) clear calibration image
     *    TEST      (root only) test calibration area
     * ------------------------------------------------------------------
     */
    case FUGA_IOC_RESET:
      retval = fuga_ccam_reset( dev );   // reset ccam to default values
      retval = fuga_reset( dev );        // write woi + flags
      break;

    case FUGA_IOC_HW_RESET:    /* reset to the i/o line to the FLEX board */
      outb(0x01, dev->amcc.address[0] + 0x3f);
      udelay( 10 );
      outb(0x00, dev->amcc.address[0] + 0x3f);
      retval = 0;
      break;
    case FUGA_IOC_CONFIG:       /* reset the flex board */
      retval = fuga_flex_config( dev); 
      break;
    case FUGA_IOC_CLEAR:
      retval = fuga_flash_clear( dev );
      break;
    case FUGA_IOC_TEST:
      retval = fuga_flash_test( dev );
      break;
    
    default:
      /*
       * unknown ioctl option
       */
      break;
   
  }
  return retval;
}

/* #####################################################################
 * MODULE
 * #####################################################################
 */

#ifdef MODULE

EXPORT_NO_SYMBOLS;

static unsigned int fuga_major = 0;       /* request major number */

const char * fuga_name = "fuga";
struct file_operations fuga_fops = {
  NULL,         /* lseek */
  fuga_read,    /* read */
  NULL,         /* write */
  NULL,         /* readdir */
  NULL,         /* select */
  fuga_ioctl,   /* ioctl */
  NULL,         /* mmap */
  fuga_open,    /* open */
  NULL,         /* flush */
  fuga_release, /* release */
  NULL,         /* fsync */
  NULL,         /* check_media_change */
  NULL,         /* revalidate */
                /* eventually nothing more */
};

int init_module(void)
{   int err = 0;

    PRINTD("FUGA: init_module \n");
    
    fuga_dev = kmalloc(sizeof(struct fuga_device), GFP_KERNEL);
    if ( fuga_dev == NULL) {
      /* error */
      PRINTD("fuga: unable to get free page Ret %p \n", fuga_dev);
      return (-1);
    }

    memset(fuga_dev, 0, sizeof(struct fuga_device) );

    if (fuga_hw_init( fuga_dev ) != 0) {
      kfree( fuga_dev );
      return (-1);
    }
    fuga_ccam_reset( fuga_dev );

    err = register_chrdev(fuga_major, fuga_name, &fuga_fops);
    if (err < 0) {
      kfree( fuga_dev );
      PRINTD("fuga register_chrdev error %d\n", err);
      return err;
    }
    if (fuga_major == 0) 
      fuga_major = err;
      PRINTZ("fuga major %d\n", fuga_major);

    return 0;
}

void cleanup_module(void)
{
    
    unregister_chrdev(fuga_major, fuga_name);
    kfree( fuga_dev );     return;
}

#endif /* defined MODULE */


#endif /* version 1.99.4 or newer */
